#pragma once

#include <queue>		// priorityqueue<>

// CAStar, V:1.2
//
// Info: simple and efficient AStar implementation, very flexible and easy to use
// Additional info: code was written with tab size 2, compiled with Visual Studio
// Author: Martin Mittring (Kosmokleaner@Kosmokleaner.de)
// Copyright: code is free to use in any way, use at your own risk
// Files: AStar.h, multiple files for the win32 sample application
// Dependencies: STL priority_queue		#include <queue>
// ToDo:
//   * improve documentation
//   * more testing

// Suggested Integration for pathfinding (object starts at a given position and should reach the given goal position as few moves as possible):
//
//   a) TAStarNodeHandle is "CNode *", it holds a reference count, the object position, heuristic, cost and a pointer to the parent
//      duplicate nodes are detected by a 2D bool array in the graph object, new and delete can cost quite some performance (AStarTestApp example)
//   b) like a) but the new and delete are avoided by [re]using an array of CNode objects, TAStarNodeHandle becomes "unsigned short/int" to index
//      CNode objects in the array
//   c) The max CNode count can be limited to the map size and therefore be fully preallocated
//   d) TAStarNodeHandle is a 2D position (int x, int y), all other data is stored in a 2D array, duplicated nodes are handled
//      naturally as the array can only hold one nodes in one position. Memory occupation is constant if limited to one AStar solver at a time.
//
//  To save memory the heuristic can be computed instead of being stored, some values can be 16 or 8 bit only. To avoid unneccessary memory
//  allocations the CAStar object should be reused by calling Reset().
//
//  More complex movements might require to store not only the absolute positions but also the movement operation (e.g. position stored in float,
//  movement in float rotation and forward movement).
//
//  The AStar algorithm can be easily adjusted by the heuristic and cost function (e.g. cheaper movement costs on streets, higher cost around
//  enemies).
//
// the closed list (to avoid processing nodes more than once) needs to be handled by the caller in ExpandNode (various implementations are possible)


template <class TAStarGraph, class TAStarNodeHandle >
class CAStar
{
public:

	//! constructor
	CAStar( TAStarGraph &rGraph ) 
		:m_OpenNodes(m_Comparator), m_Comparator(rGraph)
	{
	}

	//! useful to avoid memory allocation when using AStar multiple times
	void ResetAndSetStart( TAStarNodeHandle StartNode )
	{
		while(!m_OpenNodes.empty())			// todo: clear container directly to be more efficient (solution: derive from priority queue) 
			m_OpenNodes.pop();

		PushNode(StartNode);
	}

	//! call to evaluate one node
	TAStarNodeHandle Process()
	{
		if(m_OpenNodes.empty())
			return TAStarNodeHandle();

		TAStarNodeHandle BestNode = m_OpenNodes.top();
		m_OpenNodes.pop();

		m_Comparator.m_rGraph.ExpandNode< CAStar<TAStarGraph,TAStarNodeHandle> >(BestNode,*this);

		return BestNode;
	}

	//! should be called in ExpandNode()
	void PushNode( TAStarNodeHandle Node )
	{
		m_Comparator.m_rGraph.OnPushNode(Node);
		m_OpenNodes.push(Node);
	}

	//! useful for the client
	TAStarGraph &GetGraph() const
	{
		return m_Comparator.m_rGraph;
	}
	
private: // -----------------------------------------------------

	//! to map the node comparison to m_rGraph
	struct SComparator
	{
		SComparator( TAStarGraph &rGraph ) 
			:m_rGraph(rGraph)
		{
		}

		bool operator() ( const TAStarNodeHandle a, const TAStarNodeHandle b ) const
		{
			return m_rGraph.ComparePaths(a,b);
		}

		TAStarGraph &							m_rGraph;					//!< required by the SComparator and the parent class CAStar
	};

	// ------------------------------------------------------------

	SComparator																																					m_Comparator;				//!< to map the node comparison to m_rGraph
	std::priority_queue<TAStarNodeHandle,std::vector<TAStarNodeHandle>,SComparator >		m_OpenNodes;				//!<
};
