#pragma once

// CAStar2DGrid, V:1.2
//
// Info: <todo>
// Additional info: code was written with tab size 2, compiled with Visual Studio
// Author: Martin Mittring (Kosmokleaner@Kosmokleaner.de)
// Copyright: code is free to use in any way, use at your own risk
// Files: CAStar2DGrid.h, CAStar2DGrid.cpp, multiple files for the win32 sample application
// Dependencies: STL priority_queue		#include <queue>
//               AStar.h

#include <assert.h>								// assert()
#include "AStar.h"								// CAStar<>

class CAStar2DGrid
{
	typedef unsigned char uint8;
	typedef unsigned short uint16;
	typedef unsigned int uint32;

	// array needs to be max 256x256 because m_CostPlusHeuristic is uint8
	class CArrayNode
	{
	public:

		uint32			m_CostPlusHeuristic:13;		//!< cost + heuristic to compare nodes efficiently (redundant)
		uint32			m_Cost:13;								//!< accumulated movement cost
		int					m_ParentAndType:6;				//!< 8 neighbours, numbered like this: 123 456 789, 0 means walkable but not set, -1 means not walkable

		//! constructor
		CArrayNode() :m_ParentAndType(0), m_Cost(0), m_CostPlusHeuristic(0)
		{
		}
	};

public: // ------------------------------------------------------------

	struct SArrayPos
	{
		//! default constructor
		SArrayPos()
			:m_X(-1), m_Y(-1)
		{
		}

		//! constructor
		SArrayPos( const uint16 X, const uint16 Y )
			:m_X(X), m_Y(Y)
		{
		}

		bool operator==( const SArrayPos &v ) const
		{
			return m_X==v.m_X && m_Y==v.m_Y;
		}

		operator bool() const
		{
			return m_X>=0;
		}

		short			m_X;			//!< >=0 if a valid position, -1 if not valid
		short			m_Y;			//!< >=0 if a valid position, -1 if not valid
	};

private: // ------------------------------------------------------------

	class CAStar2DGridGraph
	{
	public:
		CAStar2DGridGraph( const uint32 dwWidth, const uint32 dwHeight );

		void SetGoal( const SArrayPos v );

		//! useful to stop execution and show goal on the map
		bool IsGoal( const SArrayPos v ) const;

		//! client can call this
		void SetObstacle( const SArrayPos v );

		//! client can call this
		void ClearObstacle( const SArrayPos v );

		//! client can call this
		bool GetObstacle( const SArrayPos v ) const;

		//! required by CAStar - to compare two nodes
		//! \return true if r is considered to be a better path than l
		bool ComparePaths( const SArrayPos l, const SArrayPos r ) const;

		//! required by CAStar
		void OnPushNode( const SArrayPos v );

		//! required by CAStar - to add all nodes to the open list that are connected with v,
		//! this is a template method to efficiently return new node back to the AStar
		template <class TSink>																		// TSink has GetGraph(), PushNode()
		void ExpandNode( SArrayPos v, TSink &rSink )
		{
			CAStar2DGridGraph &rGraph = rSink.GetGraph();		// not required but left in so compiler ensures the TSink has this method

			// parent:
			//
			// 123
			// 456
			// 789

			// try new node by moving to left,right,top or bottom (folloing the edge)
			ConsiderNewNode(v,rSink, SArrayPos(v.m_X-1,v.m_Y  ), 6 );
			ConsiderNewNode(v,rSink, SArrayPos(v.m_X+1,v.m_Y  ), 4 );
			ConsiderNewNode(v,rSink, SArrayPos(v.m_X  ,v.m_Y-1), 8 );
			ConsiderNewNode(v,rSink, SArrayPos(v.m_X  ,v.m_Y+1), 2 );
		}

		SArrayPos GetPrevious( const SArrayPos v ) const;

		//! to define which new nodes are possible
		bool IsWalkableAndNotKnown( const SArrayPos v ) const;

		//! for debugging purpose
		bool WasNodeTested( const SArrayPos v ) const;

	private: // ----------------------------------------------------------------

		SArrayPos												m_Goal;								//!< x,y position we want to reach
		uint32													m_dwWidth;						//!< >0
		uint32													m_dwHeight;						//!< >0
		std::vector<CArrayNode>					m_Nodes;							//!< [0..m_dwWidth*m_dwHeight-1] [x+y*m_dwWidth]

		// -----------------------------------------------------------------------
			
		//! useful to consider a new nodes in one code line 
		template <class TSink>
		void ConsiderNewNode( const SArrayPos ParentPos, TSink &rSink, const SArrayPos newOne, const char ParentId )
		{
			assert(ParentId>0 && ParentId<10);

			if(IsWalkableAndNotKnown(newOne))
			{
				CArrayNode &rParent = m_Nodes[ParentPos.m_X+ParentPos.m_Y*m_dwWidth];
				CArrayNode &rNewOne = m_Nodes[newOne.m_X+newOne.m_Y*m_dwWidth];

				rNewOne.m_Cost = rParent.m_Cost+1;
				assert(rNewOne.m_ParentAndType==0);
				rNewOne.m_ParentAndType=ParentId;

				rSink.PushNode(newOne);		
			}
		}

		friend class CAStar2DGrid;
	};

public: // ------------------------------------------------------------

	//! constructor
	CAStar2DGrid( const uint32 dwWidth, const uint32 dwHeight );

	//! useful to show goal on the map
	bool IsGoal( const SArrayPos v ) const { return m_GridGraph.IsGoal(v); }
	void SetGoal( const SArrayPos v ) { return m_GridGraph.SetGoal(v); }

	void ResetAndSetStart( const SArrayPos v ) { m_AStar.ResetAndSetStart(v); }

	void SetObstacle( const SArrayPos v ) { return m_GridGraph.SetObstacle(v); }
	void ClearObstacle( const SArrayPos v ) { return m_GridGraph.ClearObstacle(v); }
	bool GetObstacle( const SArrayPos v ) const { return m_GridGraph.GetObstacle(v); }

	uint32 GetWidth() const { return m_GridGraph.m_dwWidth; }
	uint32 GetHeight() const { return m_GridGraph.m_dwHeight; }

	//! example:
	//!
	//! SArrayPos p=Solve();				// if(p) can be used to test if no path was found
	//! while(p)
	//! { 
	//!   Dump(p); p=GetPrevious(p); 
	//! }
	SArrayPos Solve();

	//! use together with Solve()
	SArrayPos GetPrevious( const SArrayPos v ) const { return m_GridGraph.GetPrevious(v); }

	//! for debugging purpose
	bool WasNodeTested( const SArrayPos v ) const { return m_GridGraph.WasNodeTested(v); }

private: // ---------------------------------------------------------------------
	
	CAStar2DGridGraph												m_GridGraph;					//!<
	CAStar<CAStar2DGridGraph,SArrayPos>			m_AStar;							//!<
};
