// sample application to test the CAStar<> class

#include "stdafx.h"
#include "AStarTestApp.h"
#include "AStar.h"						// CAStar<>
#include <assert.h>						// assert()
#include <math.h>							// sqrtf()

class CNode;

class CGraph
{
public:

	//! constructor
	CGraph( const int iGoalX, const int iGoalY );

	//! to visualize the map
	void Debug( HDC hdc );

	//! useful to stop execution and show goal on the map
	bool IsGoal( const CNode &rNode ) const;

	//! required by CAStar - to compare the heuristic of two nodes
	//! \return true if r is considered to be a better path than l
	bool ComparePaths( const CNode *l, const CNode *r ) const;

	//! required by CAStar - used when instering new chilren and when inserting the start node
	void OnPushNode( CNode *pNode );

	//! required by CAStar - to add all nodes to the open list that are connected with v,
	//! this is a template method to efficiently return new nodes back to the AStar
	template <class TSink>																		// TSink has GetGraph(), PushNode()
	void ExpandNode( CNode *pNode, TSink &rSink )
	{
		CGraph &rGraph = rSink.GetGraph();

		// try new nodes by moving to left,right,top or bottom 
		ConsiderNewNode(pNode,rSink, pNode->m_iX-1,pNode->m_iY  );
		ConsiderNewNode(pNode,rSink, pNode->m_iX+1,pNode->m_iY  );
		ConsiderNewNode(pNode,rSink, pNode->m_iX  ,pNode->m_iY-1);
		ConsiderNewNode(pNode,rSink, pNode->m_iX  ,pNode->m_iY+1);
	}

private: // ---------------------------------------------------------------------
	
	int						m_iGoalX;							//!< x position we want to reach
	int						m_iGoalY;							//!< y position we want to reach
	bool					m_bKnown[10][10];			//!< to reject nodes we found already (closed list)

	//! to set the heuristic of a pushed new nodes
	void UpdateHeuristic( CNode *pNode ) const;

	//! to avoid processing of already found nodes (add to closed list)
	void MarkAsKnown( const CNode *pNode );

	//! to define which new nodes are possible
	bool IsWalkable( const int iX, const int iY ) const;

	//! to find new nodes
	bool IsWalkableAndNotKnown( const int iX, const int iY );

	//! useful to consider a new node in one code line 
	template <class TSink>
	void ConsiderNewNode( CNode *pNode, TSink &rSink, const int iX, const int iY )
	{
		if(IsWalkableAndNotKnown(iX,iY))
		{
			CNode *p = new CNode(iX,iY,pNode);		// new is a flaw, delete is missing

			rSink.PushNode(p);		
		}
	}
};



class CNode
{
private:

	int						m_iRefCount;					//!< to free nodes if not needed any more

public:

	int						m_iX;									//!< current x position, also used to reconstruct the path
	int						m_iY;									//!< current y position, also used to reconstruct the path
	float					m_fCostPlusHeuristic;	//!< cost + heuristic to compare nodes efficiently (redundant)
	float					m_fCost;							//!< accumulated movement costs
	CNode *				m_pParent;						//!< required once the goal was reached to reconstruct the path

	//! constructor
	CNode( const int iX, const int iY, CNode *pParent=0 ) 
		:m_iX(iX), m_iY(iY), m_pParent(pParent), m_iRefCount(0)
	{
		if(pParent)
			m_fCost = pParent->m_fCost+1.0f;
		else
			m_fCost = 0.0f;
	}

	void AddRefCount()
	{
		++m_iRefCount;
	}

	//! to visualize a node
	void DebugNode( HDC hdc )
	{
		for(int y=1;y<8;++y)
		for(int x=1;x<8;++x)
			SetPixel(hdc,(m_iX+1)*10+x,(m_iY+1)*10+y,0x808080);

		Sleep(100);
	}

	//! to visualize the path back to the start node
	void DebugPath( HDC hdc )
	{
		_DebugPath(this,hdc);
	}

	void Release()
	{
		if(!m_iRefCount)
			delete this;
		 else
			--m_iRefCount;
	}

private: // ----------------------------------------------------------------------

	static void _DebugPath( CNode *pThis, HDC hdc )
	{
		OutputDebugString("Path found (backward): ");	

		while(pThis)
		{
			char str[256];

			sprintf_s(str,sizeof(str),"(%d,%d) ",pThis->m_iX,pThis->m_iY);
			OutputDebugString(str);

			for(int y=1;y<8;++y)
			for(int x=1;x<8;++x)
				SetPixel(hdc,(pThis->m_iX+1)*10+x,(pThis->m_iY+1)*10+y,0x80ff80);

			pThis = pThis->m_pParent;
		}
	
		OutputDebugString("\n");	
	}
};



CGraph::CGraph( const int iGoalX, const int iGoalY ) 
	:m_iGoalX(iGoalX), m_iGoalY(iGoalY)
{
	// clear closed list
	for(int y=0;y<10;++y)
	for(int x=0;x<10;++x)
		m_bKnown[x][y]=false;
}

void CGraph::OnPushNode( CNode *pNode )
{
	if(pNode->m_pParent)
		pNode->m_pParent->AddRefCount();			// reference counting to free unused nodes later

	UpdateHeuristic(pNode);		// to compute and store heuristic
	MarkAsKnown(pNode);				// to avoid processsing already found nodes (add to closed list)
}

bool CGraph::IsWalkable( const int iX, const int iY ) const
{
	if(iX<0 || iX>=10
	|| iY<0 || iY>=10)
		return false;			// border

	if(iX>=3 && iX<=9
	&& iY>=4 && iY<=5)
		return false;			// obstacle

	return true;				// walkable
}


bool CGraph::IsWalkableAndNotKnown( const int iX, const int iY )
{
	if(!IsWalkable(iX,iY))
		return false;						// not walkable

	if(!m_bKnown[iX][iY])			// not in closed list
	{
		m_bKnown[iX][iY]=true;	// add to closed list

		return true;						// walkable and not known
	}

	return false;							// walkable but already known
}

void CGraph::MarkAsKnown( const CNode *pNode )
{
	m_bKnown[pNode->m_iX][pNode->m_iY]=true;
}


void CGraph::Debug( HDC hdc )
{
	for(int iY=-1;iY<11;++iY)
	for(int iX=-1;iX<11;++iX)
	{
		DWORD col = 0;							// border

		if(IsGoal(CNode(iX,iY)))
			col = 0x00ff00;						// goal
		else
			if(IsWalkable(iX,iY))
				continue;

		for(int y=0;y<10;++y)
		for(int x=0;x<10;++x)
			SetPixel(hdc,(iX+1)*10+x,(iY+1)*10+y,col);
	}
}


bool CGraph::ComparePaths( const CNode *l, const CNode *r ) const
{
	return l->m_fCostPlusHeuristic > r->m_fCostPlusHeuristic;
}


bool CGraph::IsGoal( const CNode &rNode ) const
{
	return rNode.m_iX==m_iGoalX 
			&& rNode.m_iY==m_iGoalY;
}


void CGraph::UpdateHeuristic( class CNode *pNode ) const
{
	float X = (float)(m_iGoalX - pNode->m_iX);
	float Y = (float)(m_iGoalY - pNode->m_iY);

	float L = max(abs(X),abs(Y))*2;		// manhatten distance

	pNode->m_fCostPlusHeuristic = pNode->m_fCost + L;
}


void Draw( HDC hdc )
{
	CGraph env(6,9);												// define goal
	CNode Start(3,1);												// define start

	Start.AddRefCount();										// add 1 to the reference count to supress the delete 

	env.Debug(hdc);													// draw map

	CAStar<CGraph,CNode *> astar(env);

	astar.ResetAndSetStart(&Start);

	while(CNode *pNext = astar.Process())		// find a good node to process
	{
		pNext->DebugNode(hdc);

		if(env.IsGoal(*pNext))
			pNext->DebugPath(hdc);							// draw path if finished but process further as there might be a better path

		pNext->Release();		// reference counting ensures only used nodes become deleted
	}
}



// --------------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------------

// the following code is only needed to get some base application to run the sample

// --------------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------------






const TCHAR *szWindowClass = "ASTARTESTAPP";			// the main window class name

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	int wmId, wmEvent;
	PAINTSTRUCT ps;
	HDC hdc;

	switch (message)
	{
	case WM_COMMAND:
		wmId    = LOWORD(wParam);
		wmEvent = HIWORD(wParam);

		switch (wmId)
		{
			case IDM_EXIT:
				DestroyWindow(hWnd);
				break;
			default:
				return DefWindowProc(hWnd, message, wParam, lParam);
		}
		break;
	case WM_PAINT:
		hdc = BeginPaint(hWnd, &ps);
		Draw(hdc);
		EndPaint(hWnd, &ps);
		break;
	case WM_DESTROY:
		PostQuitMessage(0);
		break;
	default:
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
	return 0;
}



int APIENTRY _tWinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, int nCmdShow )
{
	UNREFERENCED_PARAMETER(hPrevInstance);
	UNREFERENCED_PARAMETER(lpCmdLine);

	MSG msg;

	WNDCLASSEX wcex;

	wcex.cbSize = sizeof(WNDCLASSEX);

	wcex.style			= CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc	= WndProc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= hInstance;
	wcex.hIcon			= LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ASTARTESTAPP));
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
	wcex.lpszMenuName	= MAKEINTRESOURCE(IDC_ASTARTESTAPP);
	wcex.lpszClassName	= szWindowClass;
	wcex.hIconSm		= LoadIcon(wcex.hInstance, MAKEINTRESOURCE(IDI_SMALL));

	RegisterClassEx(&wcex);

	HWND hWnd = CreateWindow(szWindowClass, "AStarTestApp", WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInstance, NULL);

	if(!hWnd)
		return FALSE;

	ShowWindow(hWnd, nCmdShow);
	UpdateWindow(hWnd);


	while(GetMessage(&msg,NULL,0,0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	return (int)msg.wParam;
}